#
#  Copyright Bradley Lowekamp
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0.txt
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
from click.testing import CliRunner
from sitkibex.cli import cli
import os.path
import pytest


def _data_files():
    """
    A dictionary of test data file names to full paths.
    """

    data_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")

    file_dic = {}
    for f in os.listdir(data_dir):
        full_path = os.path.join(data_dir, f)
        if os.path.isfile(full_path):
            file_dic[f] = full_path

    return file_dic


data_files = _data_files()


def test_cli_help():
    runner = CliRunner()
    result = runner.invoke(cli, ["--help"])
    assert not result.exception


def test_cli_reg_help():
    runner = CliRunner()
    result = runner.invoke(cli, ["registration", "--help"])
    assert not result.exception


reg_args = [
    ["registration", data_files["panel1.nrrd"] + "@JOJO", data_files["panel2.nrrd"] + "@JOJO", "out.txt"],
    ["registration", data_files["panel1.nrrd"] + "@4", data_files["panel2.nrrd"] + "@4", "out.txt"],
    ["registration", data_files["panel1.nrrd"] + "@Ch5", data_files["panel2.nrrd"] + "@CH5", "out.txt"],
    ["registration", data_files["vpanel1.nrrd"] + "@JOJO", data_files["panel2.nrrd"] + "@CH5", "out.txt"],
    ["registration", data_files["vpanel1.nrrd"] + "@4", data_files["panel2.nrrd"] + "@CH5", "out.txt"],
]


@pytest.mark.parametrize("cli_args", reg_args)
def test_cli_reg(cli_args):
    runner = CliRunner()
    with runner.isolated_filesystem():
        result = runner.invoke(cli, cli_args)

    assert not result.exception


resample_args = [
    "resample {} {} -o test.nrrd".format(data_files["panel1.nrrd"], data_files["panel2.nrrd"]),
    "resample {} {}@Ch1 -o test.nrrd".format(data_files["panel1.nrrd"], data_files["panel2.nrrd"]),
    "resample --fusion {}@JOJO {}@JOJO -o test.nrrd".format(data_files["panel1.nrrd"], data_files["panel2.nrrd"]),
    "resample  --bin 2 --fusion --projection {}@1 {}@Ch1 -o test.png".format(
        data_files["panel1.nrrd"], data_files["panel2.nrrd"]
    ),
]


@pytest.mark.parametrize("cli_args", resample_args)
def test_cli_resample(cli_args):
    runner = CliRunner()
    with runner.isolated_filesystem():
        result = runner.invoke(cli, cli_args.split())
    assert not result.exception


def test_cli_resample_help():
    runner = CliRunner()
    result = runner.invoke(cli, ["resample", "--help"])
    assert not result.exception



#
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0.txt
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import xml.etree.ElementTree as ET


class XMLInfo:
    """
    The image metadata dictionary contains the following keys-values:
        'unit' - string denoting the physical units of the image origin,
                 and spacing.
        'times' - string denoting the time associated with the image in
                            ('%Y-%m-%d %H:%M:%S.%f' -
                            Year-month-day hour:minute:second.microsecond) format.
        'imaris_channels_information' - XML string denoting channel information.
        XML structure:

    <xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">
      <xs:element name="imaris_channels_information">
        <xs:complexType>
          <xs:sequence>
            <xs:element name="channel">
              <xs:complexType>
                  <xs:element type="xs:string" name="name"/>
                  <xs:element type="xs:string" name="description"/>
                  <xs:element type="xs:string" name="color"/>
                  <xs:element type="xs:string" name="range"/>
                  <xs:element type="xs:string" name="gamma" minOccurs="0" maxOccurs="1"/>
              </xs:complexType>
            </xs:element>
          </xs:sequence>
        </xs:complexType>
      </xs:element>
    </xs:schema>
    """

    def __init__(self, xml_string):
        self.root_element = ET.fromstring(xml_string)

    @staticmethod
    def _parse_tuple(s):
        if s is None:
            return None
        return s.replace(",", " ").split()

    @property
    def channel_names(self):
        return [e.text for e in self.root_element.findall("channel/name")]

    def descriptions(self):
        return [e.text for e in self.root_element.findall("channel/description")]

    def colors(self):
        colors = []
        for ch_element in self.root_element.iter("channel"):
            e = ch_element.find("color")
            if e is None:
                c = None
            else:
                c = [float(c) / 255 for c in self._parse_tuple(e.text)]
            colors.append(c)
        return colors

    def color_tables(self):
        color_tables = []
        for ch_element in self.root_element.iter("channel"):
            e = ch_element.find("color")
            if e is None:
                t = None
            else:
                t = [float(c) / 255 for c in self._parse_tuple(e.text)]
            color_tables.append(t)
        return color_tables

    def ranges(self):
        return [self._parse_tuple(e.text) for e in self.root_element.findall("channel/range")]

    def gamas(self):
        gamas = []
        for ch_element in self.root_element.iter("channel"):
            e = ch_element.find("gama")
            if e is None:
                g = None
            else:
                g = [float(self._parse_tuple(e.text))]
            gamas.append(g)
        return gamas

    def alpha(self):
        return [float(e.text) for e in self.root_element.findall("channel/alpha")]


class OMEInfo:

    _ome_ns = {"OME": "http://www.openmicroscopy.org/Schemas/OME/2015-01"}

    def __init__(self, ome_xml_string):
        self._root_element = ET.fromstring(ome_xml_string)

    def _image_element(self):
        return self._root_element.findall("OME:Image/OME:Pixels", self._ome_ns)[0]

    @property
    def channel_names(self):
        el = self._root_element.findall("OME:Image/OME:Pixels/OME:Channel", self._ome_ns)
        return [e.attrib["Name"] for e in el]

    @property
    def dimension_order(self):
        return self._image_element().attrib["DimensionOrder"]

    @property
    def size(self):
        img_element = self._image_element()
        size = []
        for d in self.dimension_order:
            attr = "Size{}".format(d)
            size.append(int(img_element.get(attr, 1)))
        return size

    @property
    def spacing(self):
        img_element = self._image_element()
        spacing = []
        for d in self.dimension_order:
            attr = "PhysicalSize{}".format(d)
            spacing.append(float(img_element.get(attr, 1.0)))
        return spacing

    @property
    def units(self):
        img_element = self._image_element()
        default = "µm"
        units = []
        for d in self.dimension_order:
            attr = "PhysicalSize{}Unit".format(d)
            units.append(img_element.get(attr, default))
        return units



#
#  Copyright Bradley Lowekamp
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0.txt
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import SimpleITK as sitk
import logging

_logger = logging.getLogger(__name__)


def _combine_images(fixed_image, moving_image, transform, virtual_image=None, tx2=None, fusion=False):
    """Combine two scalar images into multi-component images.

    The default operation is to compose the two input images into a 2 channel vector image.

    If fusion is True, then the images are combined into RGB.

    """

    output_pixel_type = sitk.sitkUInt8
    if (fixed_image.GetNumberOfComponentsPerPixel() > 1) or (moving_image.GetNumberOfComponentsPerPixel() > 1):
        output_pixel_type = sitk.VectorUInt8

    if virtual_image is not None:
        moving_resampled = sitk.Resample(
            moving_image, referenceImage=virtual_image, transform=transform, outputPixelType=output_pixel_type
        )
        fixed_resampled = sitk.Resample(fixed_image, virtual_image, transform=tx2, outputPixelType=output_pixel_type)
    else:
        moving_resampled = sitk.Resample(
            moving_image, referenceImage=fixed_image, transform=transform, outputPixelType=output_pixel_type
        )
        fixed_resampled = sitk.Cast(fixed_image, output_pixel_type)

    if fusion:
        moving_resampled = sitk.RescaleIntensity(moving_resampled)
        fixed_resampled = sitk.RescaleIntensity(fixed_resampled)
        return sitk.Compose([fixed_resampled, moving_resampled, fixed_resampled // 2.0 + moving_resampled // 2.0])

    return sitk.Compose([fixed_resampled, moving_resampled])


def resample(
    fixed_image: sitk.Image,
    moving_image: sitk.Image,
    transform: sitk.Transform = None,
    *,
    fusion=False,
    projection=False,
    combine=False,
    invert=False
) -> sitk.Image:
    """Resample fixed_image onto the coordinates of moving_image with transform results from registration.

    The registration process results in a transform which maps points from the coordinates of the fixed_image to the
    moving_image. This method is next used to resample the the fixed_image with the transform to produce an image with
    the fixed_image aligned with the moving_image.

    If no transform is provided, then an identity transform is assumed and the moving_image is still resampled onto the
    fixed_image. This operation is useful to see alignment of images before registration.



    :param fixed_image: A 3D SimpleITK Image whose pixel values are resampled
    :param moving_image: A 3D SimpleITK Image whose coordinates are used for the output image
    :param transform: (optional) A 3D SimpleITK Transform mapping from points from the fixed_image to the moving_image.
    :param fusion: Enable fusing the resampled moving_image and the fixed_image into a RGB image.
    :param projection: Enable perform a z-projection to reduce the dimensionality to 2D.
    :param combine: Enable combining the resampled moving_image and the fixed_image into a 2-channel vector image.
    :param invert: Invert the input transform.
    :return: The processed SimpleITK Image.
    """

    if not transform:
        transform = sitk.Transform(3, sitk.sitkIdentity)

    if invert:
        transform = transform.GetInverse()

    if fusion or combine:

        _logger.info("Fusing images...")
        resampled_image = _combine_images(fixed_image, moving_image, transform, fusion=fusion)

    else:
        output_pixel_type = moving_image.GetPixelID()

        _logger.info("Resampling image...")

        resampler = sitk.ResampleImageFilter()
        resampler.SetOutputDirection(fixed_image.GetDirection())
        resampler.SetOutputOrigin(fixed_image.GetOrigin())
        resampler.SetOutputSpacing(fixed_image.GetSpacing())
        resampler.SetSize(fixed_image.GetSize())
        resampler.SetOutputPixelType(output_pixel_type)
        resampler.SetDefaultPixelValue(0)
        resampler.SetInterpolator(sitk.sitkLinear)
        resampler.SetTransform(transform)

        resampled_image = resampler.Execute(moving_image)

    if projection:
        _logger.info("Projecting image...")
        proj_size = resampled_image.GetSize()[:2] + (0,)
        output_pixel_type = resampled_image.GetPixelID()
        projection_image = sitk.Cast(sitk.MeanProjection(resampled_image, projectionDimension=2), output_pixel_type)
        resampled_image = sitk.Extract(
            projection_image,
            size=proj_size,
            directionCollapseToStrategy=sitk.ExtractImageFilter.DIRECTIONCOLLAPSETOIDENTITY,
        )

    return resampled_image



#
#  Copyright Bradley Lowekamp
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0.txt
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
from unittest import TestCase
from sitkibex import resample
import SimpleITK as sitk
import logging


logging.basicConfig(level=logging.DEBUG)


class TestResample(TestCase):
    def test_resample1(self):
        """Test permutation of options"""

        image1 = sitk.Image([64, 64, 11], sitk.sitkFloat32)
        image2 = sitk.Image([64, 32, 12], sitk.sitkFloat32)
        image2[32, 16, 6] = 254
        image2.SetOrigin([5, 2, -1])
        tx = sitk.AffineTransform(3)
        tx.SetTranslation([5, 2, -1])

        out = resample(image1, image2)
        self.assertEqual(254, out[37, 18, 5])
        out = resample(image1, image2, tx)
        self.assertEqual(254, out[32, 16, 6])

        out = resample(image1, image2, invert=True)
        self.assertEqual(254, out[37, 18, 5])
        out = resample(image1, image2, tx, invert=True)
        self.assertEqual(254, out[42, 20, 4])

        resample(image1, image2, tx, fusion=True)

        out = resample(image1, image2, tx, combine=True)
        self.assertEqual((0, 254), out[32, 16, 6])

        out = resample(image1, image2, tx, projection=True)
        self.assertAlmostEqual(254.0 / 11.0, out[32, 16], places=5)
        resample(image1, image2, tx, fusion=True, projection=True)
        out = resample(image1, image2, tx, combine=True, projection=True)
        self.assertEqual(0, out[32, 16][0])
        self.assertEqual(int(254.0 / 11.0), out[32, 16][1])



#
#  Copyright Bradley Lowekamp
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0.txt
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import logging

default_random_seed = 100
"""Random seed used for each registration called for reproducible results. If it is set to 0 the time will be used to
initialized a seed."""

logger = logging.getLogger(__name__).parent
"""The parent `logger` object for all sub-loggers. It can be used to control the level of output and how the
 warning, info, and debug messages are handled."""

__all__ = ["default_random_seed", "logger"]



#
#  Copyright Bradley Lowekamp
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0.txt
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
import SimpleITK as sitk
import logging

_logger = logging.getLogger(__name__)


def spacing_average_magnitude(image):
    """
    Compute the log_10 magnitude to the image spacing should be scaled by to be near unit size.

    :param image: Input return.
    :return: a floating point number representing the magnitude
    """
    from math import log10

    average_spacing = sum(image.GetSpacing()) / float(image.GetDimension())
    return pow(10, round(log10(average_spacing)))


def project(image_3d, projection_func=sitk.MeanProjection):
    """ "
    Do projection from 3D to 2D along z-axis
    The direction cosine matrix is explicitly set to the identity to enable conversion of the 2-D transform back to 3D

    :param image_3d: As 3D SimpleITK Image
    :param projection_func: A SimpleITK function from the "Projection" family
    :return:
    """
    dim = 2
    size_2d = image_3d.GetSize()[:2] + (0,)

    projected_image = projection_func(image_3d, projectionDimension=dim)
    return sitk.Extract(
        projected_image, size=size_2d, directionCollapseToStrategy=sitk.ExtractImageFilter.DIRECTIONCOLLAPSETOIDENTITY
    )


def make_auto_mask(feature_image):
    """
    :param feature_image: a gray scale image
    :return: a binary ( 1's and 0's ) representing a mask of the none-zero pixels in the feature image.
    """
    return sitk.BinaryFillhole(feature_image != 0)


def fft_initialization(moving, fixed, bin_shrink=8, projection=True):
    """

    :param moving:
    :param fixed:
    :return:
    """

    if projection:
        moving = project(moving)
        fixed = project(fixed)

    try:
        moving + fixed
    except RuntimeError:
        # The addition operator will throw an exception of the images don't occupy the same space like the
        # MaskedFFTNormalizedCorrelation filter. If an exception is thrown resample onto the fixed image with the
        # identity transform. With the 3d->2d projection already making the direction matrix identity, only the origin,
        # spacing will be considered. Additionally, moth images will be the same number of pixels as required.
        resampler = sitk.ResampleImageFilter()
        resampler.SetReferenceImage(fixed)
        moving = resampler.Execute(moving)

    fraction_overlap = 0.5
    sigma = bin_shrink * fixed.GetSpacing()[0]
    pixel_type = sitk.sitkFloat32
    bin_shrink_list = [bin_shrink] * 2
    if moving.GetDimension() == 3:
        bin_shrink_list += [1]

    fft_moving = sitk.Cast(sitk.SmoothingRecursiveGaussian(sitk.BinShrink(moving, bin_shrink_list), sigma), pixel_type)
    fft_fixed = sitk.Cast(sitk.SmoothingRecursiveGaussian(sitk.BinShrink(fixed, bin_shrink_list), sigma), pixel_type)

    _logger.info("FFT Correlation...")
    out = sitk.MaskedFFTNormalizedCorrelation(
        fft_fixed,
        fft_moving,
        sitk.Cast(fft_fixed != 0, pixel_type),
        sitk.Cast(fft_moving != 0, pixel_type),
        requiredFractionOfOverlappingPixels=fraction_overlap,
    )

    _logger.info("Smoothing...")
    out = sitk.SmoothingRecursiveGaussian(out)
    _logger.info("Detecting peak...")
    _logger.info("\tConnected components and maxima...")
    cc = sitk.ConnectedComponent(sitk.RegionalMaxima(out, fullyConnected=True))
    _logger.info("\tLabel statistics...")
    stats = sitk.LabelStatisticsImageFilter()
    stats.Execute(out, cc)
    labels = sorted(stats.GetLabels(), key=lambda l_id: stats.GetMean(l_id))

    peak_idx = stats.GetBoundingBox(labels[-1])[::2]
    peak_idx = [i + 0.5 for i in peak_idx]
    peak_pt = out.TransformContinuousIndexToPhysicalPoint(peak_idx)
    peak_value = stats.GetMean(labels[-1])

    center_pt = out.TransformContinuousIndexToPhysicalPoint([p / 2.0 for p in out.GetSize()])

    translation = [c - p for c, p in zip(center_pt, peak_pt)]
    translation += [0] * (moving.GetDimension() - len(translation))

    _logger.info("FFT peak correlation of {0} at translation of {1}".format(peak_value, translation))

    return translation
